//----------------------------------------------------------------------------------------------------
/*
 *   DSESIApi.h
 *
 *
 *   Name:      dSPACE ESI API
 *
 *   Version:   3.0.1
 *
 *   Author:    dSPACE GmbH
 *
 *
 *   Copyright 2023, dSPACE GmbH. All rights reserved.
 */
//----------------------------------------------------------------------------------------------------

#ifndef DS_ESI_API_H
#define DS_ESI_API_H

#ifdef __cplusplus
    extern "C" {
#endif

#if defined _DOXYGEN || (!defined __GNUC__ && !defined _MSC_VER)
// Doxygen active --> no exports/imports
    #define DSESIAPI_DECL
    #define DSESIAPI_CONV
#else
    #if defined DSETHAPI_DEVELOPMENT
        #if defined __GNUC__
        // Exporting on Linux (no special import option)
            #define DSESIAPI_DECL __attribute__((visibility("default")))
            #define DSESIAPI_CONV
        #elif defined WIN32 || defined _MSC_VER
        // Export on Windows
            #define DSESIAPI_DECL __declspec(dllexport)
            #define DSESIAPI_CONV __stdcall
        #endif
    #else
        #if defined WIN32 || defined _MSC_VER
            #define DSESIAPI_DECL __declspec(dllimport)
            #define DSESIAPI_CONV __stdcall
        #endif
    #endif

    #ifndef DSESIAPI_DECL
        #define DSESIAPI_DECL
    #endif

    #ifndef DSESIAPI_CONV
        #define DSESIAPI_CONV
    #endif
#endif /* _DOXYGEN */

#include <stddef.h>
#include <stdint.h>

enum DSTEsiError {
    NoError,
    InternalError,
    ControlInterfaceNotAvailable,
    DataInterfaceNotAvailable,
    ChannelNotAvailable,
    InvalidChannelHandle,
    InvalidChannelIndex,
    InvalidTiming,
    InvalidPointer,
    ControlSocketError,
    DataSocketError,
    Timeout,
    InvalidInterfaceConfig,
    ChannelAlreadyRegistered,
    DataNotAvailable,
    PacketFragmentation,
};

typedef int32_t DSTEsiChannelHandle;
typedef int32_t DSTEsiChannelNumber;

typedef struct {
    // Control IP of the ESI unit
    const char* controlIP;
    // Control port of the ESI unit
    uint16_t controlPort;
    // Data IP of the ESI unit
    const char* dataIP;
    // Data base port of the ESI unit
    uint16_t dataBasePort;
    // Optional: The local ip the ESI API binds to to receive messages from the ESI unit. NULL defaults to 0.0.0.0
    const char* localIP;
    // Local port to receive messages from the ESI unit
    uint16_t localPort;
    // Maximum payload size. Is usually calculated by "MTU - (sizeof(ReplaySensorMessageHeader) + IP_HEADER_SIZE + UDP_HEADER_SIZE)"
    // Usually: sizeof(ReplaySensorMessageHeader) = 24 Byte, IP_HEADER_SIZE = 20 Byte, UDP_HEADER_SIZE = 8 Byte 
    uint16_t maxPayloadSize;
    // Timeout after which the Get functions are returning
    uint32_t timeoutGet_ms;
    // Timeout after which the ReplayStreamData function is returning (e.g. no free buffer on the ESI unit) 
    uint32_t timeoutStream_ms;
} DSTEsiInputInterfaceT;

/**
 * @brief Used to make the status of the ESI unit available.
 * 
 */
typedef struct {
    bool enabled;
    // Loss detected
    bool loss;
    // PTP is actually synchronized
    bool ptpSynced;
    // PTP activated on the ESI unit
    bool ptpActive;
    // Number of free fifo entries on the ESI unit
    uint32_t freeFifoEntries;
    // The current time on the ESI unit
    uint64_t currentTime_ns;
} DSTEsiStatusT;

/**
 * @brief Properties of a replay packet
 * 
 */
typedef struct {
    // Timestamp in nanoseconds
    uint64_t timestamp;
    // Frame number of the packet
    uint16_t frameNumber;
    // Line number of the packet
    uint16_t lineNumber;
    // Stream ID of the packet. Usually 0, change only after consulting dSPACE
    uint8_t streamID;
} DSTEsiPacketPropertiesT;

/**
 * @brief Pings the control ip and the data ip of the given inputInterface. Returns an error if one of the interfaces is not available.
 *
 * @param inputInterface IN
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_IsInputInterfaceAvailable(const DSTEsiInputInterfaceT* inputInterface);

/**
 * @brief Registers an output channel for further usage.
 *
 * @param inputInterface IN the inputInterface configuration
 * @param channelIndex IN the channel index
 * @param channelHandle OUT the channel handle which can be used to use this channel
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_RegisterOutputChannel(const DSTEsiInputInterfaceT* inputInterface, DSTEsiChannelNumber channelIndex, DSTEsiChannelHandle* channelHandle);

/**
 * @brief Unregisters an output channel from the api. MUST be called, as it frees the resources and sends a stop command to the ESI unit.
 *
 * @param channelHandle IN
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_UnregisterOutputChannel(DSTEsiChannelHandle channelHandle);

/**
 * @brief Checks if the specified output channel is available.
 *
 * @param channelHandle IN
 * @return DSTEsiError::InternalError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_IsOutputChannelAvailable(DSTEsiChannelHandle channelHandle);

/**
 * @brief Initializes the output channel for replay. Sets the specified parameters, activates replay and status messages of the ESI Unit.
 *
 * @param channelHandle IN
 * @param dataType IN
 * @param width IN
 * @param height IN
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_InitOutputChannel(DSTEsiChannelHandle channelHandle, uint16_t width, uint16_t height);

/**
 * @brief Sets the start time, the offset and optionally the stop time of the ESI Unit.
 *
 * @param channelHandle IN
 * @param StartTime IN
 * @param Offset IN
 * @param StopTime IN
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_SetReplayStartTime(DSTEsiChannelHandle channelHandle, uint64_t StartTime, uint64_t Offset = 0, uint64_t StopTime = 0);

/**
 * @brief Sets the offset of the ESI Unit
 *
 * @param channelHandle IN
 * @param Offset IN
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_SetReplayOffset(DSTEsiChannelHandle channelHandle, uint64_t Offset);
/**
 * @brief Sets the stop time of the ESI Unit.
 *
 * @param channelHandle IN
 * @param StopTime IN
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_SetReplayStopTime(DSTEsiChannelHandle channelHandle, uint64_t StopTime);

/**
 * @brief Gets the current state of the ESI Unit
 *
 * @param channelHandle IN
 * @param CurrentTime OUT
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_GetCurrentState(DSTEsiChannelHandle channelHandle, DSTEsiStatusT* Status);

/**
 * @brief Gets the current replay times of the ESI Unit
 *
 * @param channelHandle IN
 * @param StartTime OUT
 * @param Offset OUT optional
 * @param StopTime OUT optional
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_GetReplayStartTime(DSTEsiChannelHandle channelHandle, uint64_t* StartTime, uint64_t* Offset = NULL, uint64_t* StopTime = NULL);

/**
 * @brief Gets the current offset of the ESI Unit
 *
 * @param channelHandle IN
 * @param Offset OUT
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_GetReplayOffset(DSTEsiChannelHandle channelHandle, uint64_t* Offset);

/**
 * @brief Gets the current stop time of the ESI Unit
 *
 * @param channelHandle IN
 * @param StopTime OUT
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_GetReplayStopTime(DSTEsiChannelHandle channelHandle, uint64_t* StopTime);

/**
 * @brief Gets the frame size of the ESI Unit
 *
 * @param channelHandle IN
 * @param width OUT
 * @param height OUT
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_GetFrameSize(DSTEsiChannelHandle channelHandle, uint16_t* width, uint16_t* height);

/**
 * @brief Sends image data to the ESI Unit. Blocks until timeoutStream_ms is reached, if no space is available in the buffer in that time. 
 *
 * @param channelHandle IN
 * @param timestamp IN timestamp in nanoseconds
 * @param frameNumber IN the current frame number
 * @param packetProperties IN pointer to the packet properties
 * @param buffer IN pointer to the data
 * @param bufferSize IN the size of the data. If bufferSize is larger than the maximum payload size, the data is split and sent with multiple packets
 * @return DSTEsiError
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_ReplayStreamData(DSTEsiChannelHandle channelHandle, const DSTEsiPacketPropertiesT* packetProperties, const void* buffer, uint16_t bufferSize);

/**
 * @brief Resets a channel
 * 
 * @param channelHandle IN
 * @return DSTEsiError 
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_ResetChannel(DSTEsiChannelHandle channelHandle);

/**
 * @brief Clears status flags
 * 
 * @param channelHandle 
 * @return DSESIAPI_DECL 
 */
DSESIAPI_DECL DSTEsiError DSESIAPI_CONV DSESI_ClearStatusFlags(DSTEsiChannelHandle channelHandle);

/**
 * @brief Converts a given error code to a humand readable string
 *
 * @param error IN
 * @return const char*
 */
DSESIAPI_DECL const char* DSESIAPI_CONV DSESI_ErrorToString(DSTEsiError error);

#ifdef __cplusplus
    }
#endif

#endif